#include <iostream>
#include <iomanip>
#include <vector>
#include <cstdlib>

using namespace std;

#define NO_INDEX -1
#define N 100

class MyRand {
public:
	int next() { return rand() % 2; }
};

class Heapable {
public:
	int heapI;
protected:
	double eval;
public:
	virtual ~Heapable();
	double getEval() const;
	bool isLower(const Heapable *, MyRand &) const;
};

Heapable::~Heapable() {
}

double Heapable::getEval() const {
	return eval;
}

bool Heapable::isLower(const Heapable *x, MyRand &rnd) const {
	return eval < x->eval || (eval == x->eval && rnd.next() % 2 == 0);
}

class Heap {
private:
	bool valid;
	vector<Heapable *> data;
	MyRand *rnd;
public:
	Heap(MyRand &);
	void setValid(bool);
	void reset();
	void reserve(int);
	Heapable *getMin();
	bool insert(Heapable *);
	bool remove(const Heapable *);
private:
	void build();
	void heapifyUp(int i);
	void heapifyDown(int i);
	void heapifyStep(int &i, int j);
	int getLeftI(int i) const;
	int getRightI(int i) const;
	static int getUpI(int i);
	void setI(int i);
	bool testI(const Heapable *) const;
	int getMinUpI(int i) const;
	int getMinDownI(int i) const;
};

Heap::Heap(MyRand &rnd) :
	rnd(&rnd)
{
	reset();
}

void Heap::setValid(bool v) {
	if (v && !valid) {
		build();
	}
	valid = v;
}

void Heap::reset() {
	valid = true;
	data.clear();
}

void Heap::reserve(int n) {
	data.reserve(n);
}

Heapable *Heap::getMin() {
	setValid(true);
	return data.size() ? data[0] : NULL;
}

bool Heap::insert(Heapable *x) {
	if (x == NULL) {
		return false;
	}
	data.push_back(x);
	if (valid) {
		heapifyUp(data.size() - 1);
	}
	return true;
}

bool Heap::remove(const Heapable *x) {
	if (x == NULL) {
		return false;
	}
	setValid(true);
	if (!testI(x)) {
		return false;
	}
	int i = x->heapI;
	if (i == (int)data.size() - 1) {
		data.pop_back();
		return true;
	}
	data[i] = data.back();
	data.pop_back();
	if (i > 0 && getMinUpI(i) == i) {
		heapifyUp(i);
	}
	else {
		heapifyDown(i);
	}
	return true;
}

void Heap::build() {
	for (int i = (int)data.size() / 2 - 1; i >= 0; i--) {
		heapifyDown(i);
	}
	for (unsigned i = 0; i < data.size(); i++) {
		setI(i);
	}
}

void Heap::heapifyUp(int i) {
	int minI, upI;
	while ((upI = getUpI(i)) != NO_INDEX && (minI = getMinUpI(i)) == i) {
		heapifyStep(i, upI);
	}
	setI(i);
}

void Heap::heapifyDown(int i) {
	int minI;
	while ((minI = getMinDownI(i)) != i) {
		heapifyStep(i, minI);
	}
	setI(i);
}

void Heap::heapifyStep(int &i, int j) {
	Heapable *a = data[i];
	data[i] = data[j];
	data[j] = a;
	setI(i);
	i = j;
}

int Heap::getLeftI(int i) const {
	int res = 2 * i + 1;
	return res < (int)data.size() ? res : NO_INDEX;
}

int Heap::getRightI(int i) const {
	int res = 2 * i + 2;
	return res < (int)data.size() ? res : NO_INDEX;
}

int Heap::getUpI(int i) {
	return i > 0 ? (i - 1) / 2 : NO_INDEX;
}

void Heap::setI(int i) {
	data[i]->heapI = i;
}

bool Heap::testI(const Heapable *x) const {
	return x->heapI >= 0 && x->heapI < (int)data.size() && data[x->heapI] == x;
}

int Heap::getMinUpI(int i) const {
	int
		upI = getUpI(i),
		res = i;
	if (upI != NO_INDEX && data[upI]->isLower(data[res], *rnd)) {
		res = upI;
	}
	return res;
}

int Heap::getMinDownI(int i) const {
	int
		leftI = getLeftI(i),
		rightI = getRightI(i),
		res = i;
	if (leftI != NO_INDEX && data[leftI]->isLower(data[res], *rnd)) {
		res = leftI;
	}
	if (rightI != NO_INDEX && data[rightI]->isLower(data[res], *rnd)) {
		res = rightI;
	}
	return res;
}

class Test : public Heapable {
public:
	int a;
	void setEval() { eval = a; }
};

int main() {
	cout << "hula" << endl;
	vector<Test *> data;
	for (int i = 0; i < N; i++) {
		data.push_back(new Test());
		data.back()->a = 0;
	}
	for (int i = 0; i < N; i++) {
		int pos = rand() % N;
		while (data[pos]->a != 0) {
			pos = (pos + 1) % N;
		}
		data[pos]->a = i + 1;
		data[pos]->setEval();
	}
	cout << "hula" << endl;
	MyRand rnd;
	Heap heap(rnd);
	heap.setValid(false);
	for (int i = 0; i < N; i++) {
		heap.insert(data[i]);
	}
	cout << "hula" << endl;
	heap.setValid(true);
	cout << "hula" << endl;
	for (int i = 0; i < N; i++) {
		if (data[i]->a % 2 == 0) {
			if (!heap.remove(data[i])) {
				cout << "error1" << endl;
				exit(1);
			}
		}
	}
	cout << "hula" << endl;
	while (1) {
		Heapable *h = heap.getMin();
		if (h == NULL) {
			break;
		}
		cout << dynamic_cast<Test *>(h)->a << endl;
		if (!heap.remove(h)) {
			cout << "error2" << endl;
			exit(1);
		}
	}
	for (int i = 0; i < N; i++) {
		delete data[i];
	}
	cout << "hula" << endl;
	return 0;
}
