class Circuit {
private:
	string file;
	string sModel, sIns, sOuts;
	TrieStor<string> labels;
	TrieStor<int> ins, outs; // label indexes
	vector<CWin *> cwins; // circuit windows with gates
	int cwinN; // final number of circuit windows
	int gateN, gateInN, levN; // evaluation: number of gates, gate inputs and levels (delay)
	int edgeN; // only for heap reserving
	int state;
public:
	Circuit();
	Circuit(const string &file);
	~Circuit();
	const string &getFile() const;
	int getCwinN() const;
	int getGateN() const;
	int getGateInN() const;
	int getLevN() const;
	void load(const string &file);
	void makeWins(const Params &, MyRand &);
	void split(const Params &);
	void mergeAndSave(const string &fileO);
	double eval(const Params &) const;
	void printEval() const;
private:
	void clear();
	void checkState(int x) const;
	void loadOnly(const string &file);
	void loadBlif();
	void setNames();
	void setEdges();
	void setGateInN();
	void setLevN();
	void buildEdgeHeap(Heap &, const Params &, TrieStor<CWin *> &neighbors,
		TrieStor<int> &workIns, TrieStor<int> &workOuts, TrieStor<int> &workIsect) const;
	static void removeEdges(CWin *, Heap &);
	void addEdges(CWin *, Heap &, const Params &, TrieStor<CWin *> &neighbors,
		TrieStor<int> &workIns, TrieStor<int> &workOuts, TrieStor<int> &workIsect) const;
};
