#include "rusinluk_dip_incl.h"

Heap::Heap(MyRand &rnd) :
	rnd(&rnd)
{
	reset();
}

void Heap::setValid(bool v) {
	if (v && !valid) {
		build();
	}
	valid = v;
}

void Heap::reset() {
	valid = true;
	data.clear();
}

void Heap::reserve(int n) {
	data.reserve(n);
}

Heapable *Heap::getMin() {
	setValid(true);
	return data.size() ? data[0] : NULL;
}

bool Heap::insert(Heapable *x) {
	if (x == NULL) {
		return false;
	}
	data.push_back(x);
	if (valid) {
		heapifyUp(data.size() - 1);
	}
	return true;
}

bool Heap::remove(const Heapable *x) {
	if (x == NULL) {
		return false;
	}
	setValid(true);
	if (!testI(x)) {
		return false;
	}
	int i = x->heapI;
	if (i == (int)data.size() - 1) {
		data.pop_back();
		return true;
	}
	data[i] = data.back();
	data.pop_back();
	if (i > 0 && getMinUpI(i) == i) {
		heapifyUp(i);
	}
	else {
		heapifyDown(i);
	}
	return true;
}

void Heap::build() {
	for (int i = (int)data.size() / 2 - 1; i >= 0; i--) {
		heapifyDown(i);
	}
	for (unsigned i = 0; i < data.size(); i++) {
		setI(i);
	}
}

void Heap::heapifyUp(int i) {
	int minI, upI;
	while ((upI = getUpI(i)) != NO_INDEX && (minI = getMinUpI(i)) == i) {
		heapifyStep(i, upI);
	}
	setI(i);
}

void Heap::heapifyDown(int i) {
	int minI;
	while ((minI = getMinDownI(i)) != i) {
		heapifyStep(i, minI);
	}
	setI(i);
}

void Heap::heapifyStep(int &i, int j) {
	Heapable *a = data[i];
	data[i] = data[j];
	data[j] = a;
	setI(i);
	i = j;
}

int Heap::getLeftI(int i) const {
	int res = 2 * i + 1;
	return res < (int)data.size() ? res : NO_INDEX;
}

int Heap::getRightI(int i) const {
	int res = 2 * i + 2;
	return res < (int)data.size() ? res : NO_INDEX;
}

int Heap::getUpI(int i) {
	return i > 0 ? (i - 1) / 2 : NO_INDEX;
}

void Heap::setI(int i) {
	data[i]->heapI = i;
}

bool Heap::testI(const Heapable *x) const {
	return x->heapI >= 0 && x->heapI < (int)data.size() && data[x->heapI] == x;
}

int Heap::getMinUpI(int i) const {
	int
		upI = getUpI(i),
		res = i;
	if (upI != NO_INDEX && data[upI]->isLower(data[res], *rnd)) {
		res = upI;
	}
	return res;
}

int Heap::getMinDownI(int i) const {
	int
		leftI = getLeftI(i),
		rightI = getRightI(i),
		res = i;
	if (leftI != NO_INDEX && data[leftI]->isLower(data[res], *rnd)) {
		res = leftI;
	}
	if (rightI != NO_INDEX && data[rightI]->isLower(data[res], *rnd)) {
		res = rightI;
	}
	return res;
}
