#include "rusinluk_dip_incl.h"

Resynth::Resynth(const string &file) {
	if (!file.size()) {
		return;
	}
	ifstream ifs;
	ifsOpenWait(ifs, file, ios_base::in, WAIT_FOPEN, WAIT_FOPEN_C);
	string str;
	vector<string> v;
	bool readMtd = false;
	while (getFLine(ifs, str)) {
		if (!str.size()) {
			continue;
		}
		if (str == MTD_LIST) {
			readMtd = true;
			continue;
		}
		if (!readMtd) {
			commands.push_back(str);
			continue;
		}
		parseNames(str, v);
		int mtdI = methods0.size();
		for (unsigned i = 0; i < v.size(); i++) {
			int a = atoi(v[i].c_str());
			if (a >= 0 && a < (int)commands.size()) {
				methods0.push_back(a);
			}
		}
		if ((int)methods0.size() > mtdI) {
			methods0.push_back(NO_INDEX); // terminating item
			methods.push_back(mtdI);
		}
	}
}

int Resynth::getMethodN() const {
	return methods.size();
}

void Resynth::run(int methodI, const string &fileIO, const Params &pm, SharedMem &mem) const {
	if (methodI < 0 || methodI >= (int)methods.size()) {
		return;
	}
	fileTest(fileIO);
	string
		varIO = MTD_IOFILE,
		cmd;
	int i = methods[methodI];
	while (methods0[i] != NO_INDEX) {
		cmd = commands[methods0[i++]];
		replaceStr(cmd, varIO, fileIO, 0);
		if (pm.log) { logRow("Running command: " + cmd); }
		int rc = run_0(cmd, mem);
		if (rc) {
			if (pm.logW) {
				logRow("Warning: The following command ended with return code " + toString(rc) + ": " + cmd);
			}
			if (!pm.force) {
				break;
			}
		}
	}
}

int Resynth::run_0(const string &cmd0, SharedMem &mem) {
	int
		procN = mpiProcN(),
		rank = mpiRank();
	if (rank != BOSS_RANK || procN == 1) {
		return system(cmd0.c_str());
	}
	string
		jobFile = BOSS_JOB_FILE,
		cmd = "/bin/sh -c \"" + cmd0 + " ; echo $? > " + jobFile + "\" &";
	remove(jobFile.c_str());
	system(cmd.c_str());
	ifstream ifs;
	double time0 = MPI_Wtime();
	while (1) {
		if (MPI_Wtime() - time0 >= BOSS_JOB_WAIT) {
			ifs.open(jobFile.c_str(), ios_base::in);
			if (ifs.is_open()) {
				break;
			}
			time0 = MPI_Wtime();
		}
		mem.use(); // providing any RMA operations
	}
	string str;
	getFLine(ifs, str);
	int rc = atoi(str.c_str());
	ifs.close();
	remove(jobFile.c_str());
	return rc;
}
