#include "rusinluk_dip_incl.h"

Storable::Storable(int bufN) {
	buf = NULL;
	allocBuf_0(bufN);
}

Storable::~Storable() {
	free(buf);
}

void *Storable::getBuf() const {
	return buf;
}

int Storable::getBufN() const {
	return bufN;
}

void Storable::load() {
	void *p = buf;
	load_0();
	buf = p;
	setPtrs(); // due to possibly damaged val of them (when using memcpy)
}

void Storable::send(int dest, int tag) {
	save();
	MPI_Request req;
	int a = MPI_Isend(buf, bufN, MPI_BYTE, dest, tag, MPI_COMM_WORLD, &req);
	if (a != MPI_SUCCESS) {
		logErr("Message sending error (" + toString(dest) + ", " + toString(a) + ")");
	}
	int flag = 0;
	MPI_Status status;
	do {
		MPI_Test(&req, &flag, &status);
	}
		while (!flag);
}

void Storable::recv(int src, int tag, MPI_Status &status) {
	int n = bufN;
	recv_0(src, tag, status);
	if (bufN > n) {
		allocBuf(bufN);
		recv_0(src, tag, status);
	}
}

void Storable::rmaPut(SharedMem &mem) {
	save();
	mem.lock();
	int a = MPI_Put(buf, bufN, MPI_BYTE, BOSS_RANK, 0, bufN, MPI_BYTE, mem.win);
	if (a != MPI_SUCCESS) {
		logErr("RMA sending error (" + toString(a) + ")");
	}
	mem.unlock();
}

void Storable::rmaGet(SharedMem &mem) {
	int n = bufN;
	rmaGet_0(mem);
	if (bufN > n) {
		allocBuf(bufN);
		rmaGet_0(mem);
	}
}

void Storable::allocBuf(int bufN) {
	allocBuf_0(bufN);
	setPtrs();
}

void Storable::allocBuf_0(int bufN) {
	free(buf);
	buf = (bufN > 0 ? malloc(bufN) : NULL);
	this->bufN = bufN;
}

void Storable::recv_0(int src, int tag, MPI_Status &status) {
	int a = MPI_Recv(buf, bufN, MPI_BYTE, src, tag, MPI_COMM_WORLD, &status);
	if (a != MPI_SUCCESS) {
		logErr("Message receiving error (" + toString(a) + ")");
	}
	load();
}

void Storable::rmaGet_0(SharedMem &mem) {
	mem.lock();
	int a = MPI_Get(buf, bufN, MPI_BYTE, BOSS_RANK, 0, bufN, MPI_BYTE, mem.win);
	if (a != MPI_SUCCESS) {
		logErr("RMA receiving error (" + toString(a) + ")");
	}
	mem.unlock();
	load();
}
